local VERSION = tonumber(("$Revision: 40433 $"):match("%d+"))

local PitBull = PitBull
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-06-17 20:55:24 -0400 (Sun, 17 Jun 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local _,playerClass = UnitClass("player")

local L = PitBull:L("PitBull-ABF")

local PitBull_Aura = nil

local DogTag = AceLibrary("DogTag-1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")

local newFrame = PitBull.newFrame
local new, del = PitBull.new, PitBull.del

local ABF = {}
PitBull.ABF = ABF

ABF.extraFrames = {}

local ABF_db = PitBull:AcquireDBNamespace("Layout-ABF")
ABF.db = ABF_db
PitBull:RegisterDefaults("Layout-ABF", "profile", {
	["**"] = {
		portraitSide = "left",
		auraSide = "bottom",
		--note, these are only used when debuffs are split
		buffSide = "bottom",
		debuffSide = "right",
		
		showBlankSpace = false,
		
		texts = {
			['**'] = {
				custom = "",
				style = "Custom",
				styleType = "none",
				position = "frame-center",
				size = 1,
			},
			Class = {
				custom = "[Classification] [Level:DifficultyColor] [SmartClass:ClassColor] [DruidForm:Paren] [SmartRace]",
				style = "Standard",
				styleType = "class",
				position = "powerBar-left",
			},
			Health = {
				custom = "[CurHP]/[MaxHP]",
				style = "Smart",
				styleType = "health",
				position = "healthBar-right",
			},
			Name = {
				custom = "[Name]",
				style = "Standard",
				styleType = "name",
				position = "healthBar-left",
			},
			Power = {
				custom = "[CurMP]/[MaxMP]",
				style = "Absolute",
				styleType = "power",
				position = "powerBar-right",
			},
			Combo = {
				style = "Standard",
				styleType = "combo",
				position = "frame-outright",
			},
			Experience = {
				style = "Standard",
				styleType = "experience",
				position = "expBar-center",
			},
			Reputation = {
				style = "Standard",
				styleType = "reputation",
				position = "repBar-center",
			},
			Threat = {
				custom = "[Threat]/[MaxThreat]",
				style = "Absolute Short",
				styleType = "threat",
				position = "threatBar-right",
			},
			combatText = {
				position = "frame-center",
			}
		},
		
		iconPositions = {
			['*'] = "frame-edgebottomleft",
			combatIcon = "frame-edgebottomleft",
			restIcon = "frame-edgebottomleft",
			pvpIcon = "frame-edgetopright",
			leaderIcon = "frame-edgetopleft",
			masterIcon = "frame-edgetopleft",
			raidTargetIcon = "frame-edgetop",
		},
		
		iconSizes = {
			['*'] = 1,
		},
		
		heights = {
			['*'] = 2,
			healthBar = 5,
			powerBar = 4,
			castBar = 2,
			repBar = 2,
			expBar = 2,
			threatBar = 2,
			portrait = 2,
		},
		
		barPositions = {
			portrait = 1,
			healthBar = 2,
			powerBar = 3,
			expBar = 4,
			repBar = 5,
			threatBar = 6,
			blankSpace = 7,
			castBar = 8,
		},
		
		druidManaBarHeight = 0.25,
	},
	player = {
		texts = {
			Health = {
				style = "Informational",
			}
		}
	},
	pet = {
		texts = {
			Health = {
				style = "Informational",
			}
		}
	},
	target = {
		portraitSide = "right",
		auraSide = "right",
		comboSide = "bottom",
		
		texts = {
			Health = {
				style = "Informational",
			}
		}
	}
})

local iconPriorities = {
	-- this dictates the order which icons are laid out in.
	"leaderIcon",
	"masterIcon",
	"raidTargetIcon",
	"restIcon",
	"pvpIcon",
	"combatIcon",
}
for i,v in ipairs(iconPriorities) do
	iconPriorities[i] = nil
	iconPriorities[v] = i
end

local function iconPriority_sort(alpha, bravo)
	return iconPriorities[alpha] < iconPriorities[bravo]
end

function ABF.extraFrames.blankSpace(unit, frame)
	if not ABF_db.profile[frame.group].showBlankSpace then
		return
	end
	frame.blankSpace = newFrame("Frame", frame)
end

local barList = {
	blankSpace = true,
	healthBar = true,
	powerBar = true,
	castBar = true,
	repBar = true,
	expBar = true,
	threatBar = true,
	portrait = true,
}
ABF.heights = barList

local function figureHeight(multiplier, frame)
	local num = 0
	local db = ABF_db.profile[frame.group].heights
	for k in pairs(barList) do
		if frame[k] and frame[k]:IsShown() then
			if k == "portrait" then
				local portraitSide = ABF_db.profile[frame.group].portraitSide
				if portraitSide == "top" then
					num = num + db[k]
				end
			else
				num = num + db[k]
			end
		end
	end
	return db[multiplier] * frame:GetHeight() / num
end
ABF.figureHeight = figureHeight

local function barPosition(id)
	return function(unit, frame)
		local db = ABF_db.profile[frame.group]
		local barPositions = db.barPositions
		local tmp = new()
		for k,v in pairs(barPositions) do
			if tmp[v] then
				v = 1
				while tmp[v] do
					v = v + 1
				end
				barPositions[k] = v
			end
			tmp[v] = v
		end
		tmp = del(tmp)
		local portraitSide = frame.portrait and ABF_db.profile[frame.group].portraitSide
		if portraitSide ~= "left" and portraitSide ~= "right" then
			portraitSide = nil
		end
		local current_barPosition = barPositions[id]
		local bestAttach
		local bestAttach_num
		for k,v in pairs(barPositions) do
			if v < current_barPosition and frame[k] and (k ~= "portrait" or not portraitSide) then
				if not bestAttach_num or bestAttach_num < v then
					bestAttach = k
					bestAttach_num = v
				end
			end
		end
		local height = figureHeight(id, frame)
		if id == "powerBar" then
			if frame.druidManaBar then
				height = height * (1 - ABF_db.profile[frame.group].druidManaBarHeight)
			end
		end
		if bestAttach then
			if bestAttach == "powerBar" and frame.druidManaBar then
				bestAttach = 'druidManaBar'
			end
			frame[id]:SetPoint("TOP", frame[bestAttach], "BOTTOM", 0, -1)
			frame[id]:SetHeight(height - 1)
		else
			frame[id]:SetPoint("TOP", frame, "TOP")
			frame[id]:SetHeight(height) -- no pixel spacing, there is always 1 spacing less than bars.
		end
		
		if portraitSide == "left" then
			frame[id]:SetPoint("LEFT", frame.portrait, "RIGHT", 2, 0)
		else
			frame[id]:SetPoint("LEFT", frame, "LEFT")
		end
		if portraitSide == "right" then
			frame[id]:SetPoint("RIGHT", frame.portrait, "LEFT", -2, 0)
		else
			frame[id]:SetPoint("RIGHT", frame, "RIGHT")
		end
	end
end

local blankSpaceFunc = barPosition('blankSpace')
ABF.blankSpace = function(unit, frame)
	return blankSpaceFunc(unit, frame)
end
ABF.healthBar = barPosition('healthBar')
ABF.powerBar = barPosition('powerBar')
ABF.threatBar = barPosition('threatBar')
if playerClass == "DRUID" then
	function ABF.druidManaBar(unit, frame)
		if not frame.powerBar then
			return
		end
		local height = figureHeight('powerBar', frame) * ABF_db.profile[frame.group].druidManaBarHeight
		frame.druidManaBar:SetPoint("TOPLEFT", frame.powerBar, "BOTTOMLEFT")
		frame.druidManaBar:SetPoint("TOPRIGHT", frame.powerBar, "TOPRIGHT")
		frame.druidManaBar:SetHeight(height-1)
	end
end
local castBarFunc_normal = barPosition('castBar')
ABF.repBar = barPosition('repBar')
ABF.expBar = barPosition('expBar')
function ABF.castBar(unit, frame)
	castBarFunc_normal(unit, frame)
	local height = frame.castBar:GetHeight()
	frame.castBar.bg:ClearAllPoints()
	frame.castBar.bg:SetHeight(height)
	for i = 1, frame.castBar:GetNumPoints() do
		frame.castBar.bg:SetPoint(frame.castBar:GetPoint(i))
		if frame.castBarIcon then
			local point, attach, relpoint, x, y = frame.castBar:GetPoint(i)
			if point == "LEFT" then
				frame.castBar:SetPoint(point, attach, relpoint, x + height, y)
			end
		end
	end
end
function ABF.castBarIcon(unit, frame)
	local height = figureHeight('castBar', frame)
	frame.castBarIcon:SetHeight( height -1)
	frame.castBarIcon:SetWidth( height -1)
	frame.castBarIcon:SetPoint("RIGHT", frame.castBar, "LEFT" )
	frame.castBarIcon:SetTexCoord(0.07, 0.93, 0.07, 0.93)
end
function ABF.castBarText(unit, frame)
	frame.castBarText:ClearAllPoints()
	frame.castBarText:SetAllPoints(frame.castBar)
	frame.castBarText:SetJustifyH("LEFT")
end
function ABF.castBarTimeText(unit, frame)
	frame.castBarTimeText:ClearAllPoints()
	frame.castBarTimeText:SetAllPoints(frame.castBar)
	frame.castBarTimeText:SetJustifyH("RIGHT")
end

local portraitFunc_normal = barPosition('portrait')
function ABF.portrait(unit, frame)
	local portraitSide = ABF_db.profile[frame.group].portraitSide
	local leftOrRight = portraitSide == "left" or portraitSide == "right"
	if not leftOrRight then
		return portraitFunc_normal(unit, frame)
	end
	
	if portraitSide == "left" then
		frame.portrait:SetPoint("LEFT", frame, "LEFT")
	else -- right
		frame.portrait:SetPoint("RIGHT", frame, "RIGHT", 0, 0)
	end
	
	local height = frame:GetHeight()
	
	frame.portrait:SetWidth(height)
	frame.portrait:SetHeight(height)
end
if playerClass == "ROGUE" or playerClass == "DRUID" then
	function ABF.comboPoints(unit, frame)
		if frame.group ~= "target" then return end 
		local side = ABF_db.profile[frame.group].comboSide
		if side == "left" or side == "right" then
			if side == "left" then
				frame.comboPoints:SetPoint( "RIGHT", frame, "LEFT", 0, 0)
			else
				frame.comboPoints:SetPoint( "LEFT", frame, "RIGHT", 0, 0)
			end
			frame.comboPoints:SetWidth( 10 )
			frame.comboPoints:SetHeight( 50 )
			for i = 1, 5 do
				local combo = frame.comboPoints["combo"..i]
				combo:ClearAllPoints()
				if i == 1 then
					combo:SetPoint("BOTTOM", frame.comboPoints, "BOTTOM")
				else
					combo:SetPoint("BOTTOM", frame.comboPoints["combo"..(i-1)], "TOP", 0, 2)
				end			
			end
		elseif side == "top" or side == "bottom" then
			if side == "top" then
				frame.comboPoints:SetPoint( "TOPRIGHT", frame, "TOPRIGHT", 0, 0)
			else
				frame.comboPoints:SetPoint( "BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
			end
			frame.comboPoints:SetWidth( 50 )
			frame.comboPoints:SetHeight( 10 )
			for i = 1, 5 do
				local combo = frame.comboPoints["combo"..i]
				combo:ClearAllPoints()
				if i == 1 then
					combo:SetPoint("RIGHT", frame.comboPoints, "RIGHT")
				else
					combo:SetPoint("RIGHT", frame.comboPoints["combo"..(i-1)], "LEFT")
				end	
			end
		end
	end
end

function ABF.auraFrame(unit, frame)
	frame.auraFrame:ClearAllPoints()
	local size = PitBull:GetModule("Aura").db.profile[frame.group].buffsize
	frame.auraFrame:SetHeight(size)
	if PitBull:GetModule("Aura").db.profile[frame.group].split then
		local buffside = ABF_db.profile[frame.group].buffSide
		local w = frame:GetWidth()
		frame.auraFrame:SetWidth(w)
		if buffside == "bottom" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
		elseif buffside == "top" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, 2)
		elseif buffside == "left" then
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "LEFT", w-size-2, -2)
		elseif buffside == "right" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
		elseif buffside == "topleft" then
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "TOPLEFT", w-size-2, 0)
		elseif buffside == "topright" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
		elseif buffside == "bottomleft" then
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", w-size-2, 0)
		elseif buffside == "bottomright" then
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
		end
	else
		local side = ABF_db.profile[frame.group].auraSide
		if side == "bottom" then
			frame.auraFrame:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", 2, -2)
		elseif side == "top" then
			local split = PitBull:GetModule("Aura").db.profile[frame.group].split
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, split and 18 or 2)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 2, split and 18 or 2)
		elseif side == "left" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "LEFT", -2, 0)
		elseif side == "topright" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
		elseif side == "bottomright" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
		elseif side == "bottomleft" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", -2, 0)
		elseif side == "topleft" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("TOPRIGHT", frame, "TOPLEFT", -2, 0)
		elseif side == "right" then
			frame.auraFrame:SetWidth(160)
			frame.auraFrame:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
		end
	end
end
function ABF.auraFrame2(unit, frame)
	frame.auraFrame2:ClearAllPoints()
	local size = PitBull:GetModule("Aura").db.profile[frame.group].debuffsize
	local w = frame:GetWidth()
	frame.auraFrame2:SetHeight(size)
	frame.auraFrame2:SetWidth(w)
	local debuffside = ABF_db.profile[frame.group].debuffSide
	if debuffside == "bottom" then
		frame.auraFrame2:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", -2, -2)
	elseif debuffside == "top" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", -2, 2)
	elseif debuffside == "left" then
		frame.auraFrame2:SetPoint("BOTTOMRIGHT", frame, "LEFT", w-size-2, -2)
	elseif debuffside == "right" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "RIGHT", 2, 0)
	elseif debuffside == "topleft" then
		frame.auraFrame2:SetPoint("TOPRIGHT", frame, "TOPLEFT", w-size-2, 0)
	elseif debuffside == "topright" then
		frame.auraFrame2:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, 0)
	elseif debuffside == "bottomleft" then
		frame.auraFrame2:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", w-size-2, 0)
	elseif debuffside == "bottomright" then
		frame.auraFrame2:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 0)
	end
end

local function findTextAtPosition(group, position, frame)
	for k, v in pairs(ABF_db.profile[group].texts) do
		if v.position == position then
			local id = "layoutText" .. k
			if not frame or frame[id] then
				return id
			end
			if frame[k] then
				return k
			end
			return nil
		end
	end
end

local function findIconsAtPosition(group, position, frame)
	local t = new()
	for k, v in pairs(ABF_db.profile[group].iconPositions) do
		if v == position and frame[k] then
			t[#t+1] = k
		end
	end
	table.sort(t, iconPriority_sort)
	return t
end

local function findIconAtPositionBeforeIcon(group, position, frame, icon)
	local t = findIconsAtPosition(group, position, frame)
	for i,v in ipairs(t) do
		if v == icon then
			local preicon = t[i-1]
			t = del(t)
			return preicon
		end
	end
	t = del(t)
	return nil
end

local function findLastIconAtPosition(group, position, frame)
	local t = findIconsAtPosition(group, position, frame)
	for i = #t, 1, -1 do
		local icon = t[i]
		t = del(t)
		return icon
	end
	t = del(t)
	return nil
end

local function findFirstIconAtPosition(group, position, frame)
	local t = findIconsAtPosition(group, position, frame)
	for i = 1, #t do
		local icon = t[i]
		t = del(t)
		return icon
	end
	t = del(t)
	return nil
end

local function findTextOrIconAtPosition(group, position, frame, direction)
	local attach, point = ("-"):split(position)
	local approach
	if attach == "frame" then
		if point == "outtopleft" or point == "outtop" or point == "outbottomleft" or point == "outbottom" or point == "outrighttop" or point == "outright" or point == "outrightbottom" or point == "center" or point == "intopleft" or point == "intop" or point == "inbottomleft" or point == "inbottom" or point == "inleft" or point == "edgetopleft" or point == "edgetop" or point == "edgeleft" or point == "edgebottomleft" or point == "edgebottom" then
			approach = direction == "left" and "left-first" or "right-last"
		else
			approach = direction == "left" and "left-last" or "right-first"
		end
	else
		if point == "left" or point == "center" or point == "outright" then
			approach = direction == "left" and "left-first" or "right-last"
		else
			approach = direction == "left" and "left-last" or "right-first"
		end
	end
	if approach == "left-first" then
		return findFirstIconAtPosition(group, position, frame) or findTextAtPosition(group, position, frame)
	elseif approach == "left-last" then
		return findLastIconAtPosition(group, position, frame) or findTextAtPosition(group, position, frame)
	elseif approach == "right-first" then
		return findTextAtPosition(group, position, frame) or findFirstIconAtPosition(group, position, frame)
	else--if approach == "right-last" then
		return findTextAtPosition(group, position, frame) or findLastIconAtPosition(group, position, frame)
	end
end

local function textPosition(id, name)
	return function(unit, frame)
		local group = frame.group
		local position = ABF_db.profile[group].texts[name].position or ""
		if position == "hide" then
			return
		end
		local attach, point = ("-"):split(position)
		if attach and point then
			if attach == "frame" then
				if point == "outtopleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMLEFT", frame[icon], "BOTTOMRIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", 2, 5)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("BOTTOM")
					local center = findTextOrIconAtPosition(group, "frame-outtop", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, "frame-outtopright", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "outtopright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMRIGHT", frame[icon], "BOTTOMLEFT", -3, 0)
					else
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -2, 5)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("BOTTOM")
					local center = findTextOrIconAtPosition(group, "frame-outtop", frame, "right")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "outtop" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMLEFT", frame[icon], "BOTTOMRIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOM", frame, "TOP", 0, 5)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("BOTTOM")
				elseif point == "outbottomleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPLEFT", frame[icon], "TOPRIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOPLEFT", frame, "TOPLEFT", 2, -5)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("TOP")
					local center = findTextOrIconAtPosition(group, "frame-outbottom", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, "frame-outbottomright", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "outbottomright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPRIGHT", frame[icon], "TOPLEFT", -3, 0)
					else
						frame[id]:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", -2, -5)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("TOP")
					local center = findTextOrIconAtPosition(group, "frame-outbottom", frame, "right")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "outbottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPLEFT", frame[icon], "TOPRIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOP", frame, "BOTTOM", 0, -5)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("TOP")
				elseif point == "outlefttop" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPRIGHT", frame[icon], "TOPLEFT", -3, 0)
					else
						frame[id]:SetPoint("TOPRIGHT", frame, "TOPLEFT", -5, -2)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("TOP")
				elseif point == "outleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame, "LEFT", -5, 0)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "outleftbottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMRIGHT", frame[icon], "BOTTOMLEFT", -3, 0)
					else
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", -5, -2)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("BOTTOM")
				elseif point == "outrighttop" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPLEFT", frame[icon], "TOPRIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOPLEFT", frame, "TOPRIGHT", 5, -2)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("TOP")
				elseif point == "outright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame, "RIGHT", 5, 0)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "outrightbottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMLEFT", frame[icon], "BOTTOMRIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 5, 2)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("BOTTOM")
				elseif point == "intopleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPLEFT", frame[icon], "TOPRIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOPLEFT", frame, "TOPLEFT", 2, -5)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("TOP")
					local center = findTextOrIconAtPosition(group, "frame-intop", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, "frame-intopright", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "intopright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPRIGHT", frame[icon], "TOPLEFT", -3, 0)
					else
						frame[id]:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -2, -5)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("TOP")
					local center = findTextOrIconAtPosition(group, "frame-intop", frame, "right")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "intop" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("TOPLEFT", frame[icon], "TOPRIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOP", frame, "TOP", 2, -5)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("TOP")
				elseif point == "inbottomleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMLEFT", frame[icon], "BOTTOMRIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 2, 5)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("BOTTOM")
					local center = findTextOrIconAtPosition(group, "frame-inbottom", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, "frame-inbottomright", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "inbottomright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMRIGHT", frame[icon], "BOTTOMLEFT", -3, 0)
					else
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -2, 5)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("BOTTOM")
					local center = findTextOrIconAtPosition(group, "frame-inbottom", frame, "right")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "inbottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("BOTTOMLEFT", frame[icon], "BOTTOMRIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOM", frame, "BOTTOM", 2, 5)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("BOTTOM")
				elseif point == "inleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame, "LEFT", 5, 0)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("MIDDLE")
					local center = findTextOrIconAtPosition(group, "frame-center", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, "frame-inright", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "inright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame, "RIGHT", -5, 0)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("MIDDLE")
					local center = findTextOrIconAtPosition(group, "frame-center", frame, "right")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "center" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("CENTER", frame, "CENTER", 0, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgetopleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame, "TOPLEFT", -5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgetopright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame, "TOPRIGHT", 5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgetop" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("CENTER", frame, "TOP", 0, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgebottomleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame, "BOTTOMLEFT", -5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgebottomright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame, "BOTTOMRIGHT", 5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgebottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("CENTER", frame, "BOTTOM", 0, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgeleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame, "LEFT", -5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "edgeright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame, "RIGHT", 5, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				end
			elseif frame[attach] then
				if point == "left" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame[attach], "LEFT", 3, 0)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("MIDDLE")
					local center = findTextOrIconAtPosition(group, attach .. "-center", frame, "left")
					if center then
						frame[id]:SetPoint("RIGHT", frame[center], "LEFT", -3, 0)
					else
						local right = findTextOrIconAtPosition(group, attach .. "-right", frame, "left")
						if right then
							frame[id]:SetPoint("RIGHT", frame[right], "LEFT", -3, 0)
						end
					end
				elseif point == "center" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("CENTER", frame[attach], "CENTER", 0, 0)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "right" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame[attach], "RIGHT", -3, 0)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("MIDDLE")
					local center = findTextOrIconAtPosition(group, attach .. "-center", frame, "left")
					if center then
						frame[id]:SetPoint("LEFT", frame[center], "RIGHT", 3, 0)
					end
				elseif point == "top" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("TOP", frame[attach], "TOP", 0, -3)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("TOP")
				elseif point == "bottom" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("BOTTOM", frame[attach], "BOTTOM", 0, 3)
					end
					frame[id]:SetJustifyH("CENTER")
					frame[id]:SetJustifyV("BOTTOM")
				elseif point == "outright" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
					else
						frame[id]:SetPoint("LEFT", frame[attach], "RIGHT", 3, 0)
					end
					frame[id]:SetJustifyH("LEFT")
					frame[id]:SetJustifyV("MIDDLE")
				elseif point == "outleft" then
					local icon = findLastIconAtPosition(group, position, frame)
					if icon then
						frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
					else
						frame[id]:SetPoint("RIGHT", frame[attach], "LEFT", -3, 0)
					end
					frame[id]:SetJustifyH("RIGHT")
					frame[id]:SetJustifyV("MIDDLE")
				end
			end
		end
	end
end

local function getHalfWidth(group, position, frame)
	local num = 0
	if findTextAtPosition(group, position, frame) then
		num = num + 20 -- assumes a width of 40
	end
	local icons = findIconsAtPosition(group, position, frame)
	local iconSizes = ABF_db.profile[group].iconSizes
	for _,v in ipairs(icons) do
		num = num + 7.5 * iconSizes[icons]
	end
	icons = del(icons)
	return -num
end

local function iconPosition(id)
	return function(unit, frame)
		local group = frame.group
		local position = ABF_db.profile[group].iconPositions[id]
		if position == "hide" then
			return
		end
		local size = ABF_db.profile[group].iconSizes[id] * 15
		frame[id]:SetWidth(size)
		frame[id]:SetHeight(size)
		local attach, point = ("-"):split(position)
		if attach and point then
			local icon = findIconAtPositionBeforeIcon(group, position, frame, id)
			if icon then
				local left = false
				if attach == "frame" then
					if point == "outtopleft" or point == "outtop" or point == "outbottomleft" or point == "outbottom" or point == "outrighttop" or point == "outright" or point == "outrightbottom" or point == "center" or point == "intopleft" or point == "intop" or point == "inbottomleft" or point == "inbottom" or point == "inleft" or point == "edgetopleft" or point == "edgetop" or point == "edgeleft" or point == "edgebottomleft" or point == "edgebottom" then
						left = true
					end
				else
					if point == "left" or point == "center" or point == "outright" then
						left = true
					end
				end
				if left then
					frame[id]:SetPoint("LEFT", frame[icon], "RIGHT", 3, 0)
				else
					frame[id]:SetPoint("RIGHT", frame[icon], "LEFT", -3, 0)
				end
			else
				if attach == "frame" then
					if point == "outtopleft" then
						frame[id]:SetPoint("BOTTOMLEFT", frame, "TOPLEFT", 2, 5)
					elseif point == "outtopright" then
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -2, 5)
					elseif point == "outtop" then
						frame[id]:SetPoint("BOTTOMLEFT", frame, "TOP", getHalfWidth(group, position, frame), 5)
					elseif point == "outbottomleft" then
						frame[id]:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", 2, -5)
					elseif point == "outbottomright" then
						frame[id]:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", -2, -5)
					elseif point == "outbottom" then
						frame[id]:SetPoint("TOPLEFT", frame, "BOTTOM", getHalfWidth(group, position, frame), -5)
					elseif point == "outlefttop" then
						frame[id]:SetPoint("TOPRIGHT", frame, "TOPLEFT", -5, -2)
					elseif point == "outleft" then
						frame[id]:SetPoint("RIGHT", frame, "LEFT", -5, 0)
					elseif point == "outleftbottom" then
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", -5, 2)
					elseif point == "outrighttop" then
						frame[id]:SetPoint("TOPLEFT", frame, "TOPRIGHT", 5, -2)
					elseif point == "outright" then
						frame[id]:SetPoint("LEFT", frame, "RIGHT", 5, 0)
					elseif point == "outrightbottom" then
						frame[id]:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 5, 2)
					elseif point == "center" then
						frame[id]:SetPoint("LEFT", frame, "CENTER", getHalfWidth(group, position, frame), 0)
					elseif point == "intopleft" then
						frame[id]:SetPoint("TOPLEFT", frame, "TOPLEFT", 2, -5)
					elseif point == "intop" then
						frame[id]:SetPoint("TOPLEFT", frame, "TOP", getHalfWidth(group, position, frame), -5)
					elseif point == "intopright" then
						frame[id]:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -2, -5)
					elseif point == "inbottomleft" then
						frame[id]:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 2, 5)
					elseif point == "inbottom" then
						frame[id]:SetPoint("BOTTOMLEFT", frame, "BOTTOM", getHalfWidth(group, position, frame), 5)
					elseif point == "inbottomright" then
						frame[id]:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -2, 5)
					elseif point == "inleft" then
						frame[id]:SetPoint("LEFT", frame, "LEFT", 2, 0)
					elseif point == "inright" then
						frame[id]:SetPoint("RIGHT", frame, "RIGHT", -2, 0)
					elseif point == "edgetopleft" then
						frame[id]:SetPoint("CENTER", frame, "TOPLEFT", 0, 0)
					elseif point == "edgetop" then
						frame[id]:SetPoint("LEFT", frame, "TOP", getHalfWidth(group, position, frame), 0)
					elseif point == "edgetopright" then
						frame[id]:SetPoint("CENTER", frame, "TOPRIGHT", 0, 0)
					elseif point == "edgeleft" then
						frame[id]:SetPoint("CENTER", frame, "LEFT", 0, 0)
					elseif point == "edgeright" then
						frame[id]:SetPoint("CENTER", frame, "RIGHT", 0, 0)
					elseif point == "edgebottomleft" then
						frame[id]:SetPoint("CENTER", frame, "BOTTOMLEFT", 0, 0)
					elseif point == "edgebottom" then
						frame[id]:SetPoint("LEFT", frame, "BOTTOM", getHalfWidth(group, position, frame), 0)
					elseif point == "edgebottomright" then
						frame[id]:SetPoint("CENTER", frame, "BOTTOMRIGHT", 0, 0)
					end
				elseif frame[attach] then
					if point == "left" then
						frame[id]:SetPoint("LEFT", frame[attach], "LEFT", 3, 0)
					elseif point == "center" then
						frame[id]:SetPoint("LEFT", frame[attach], "CENTER", getHalfWidth(group, position, frame), 0)
					elseif point == "right" then
						frame[id]:SetPoint("RIGHT", frame[attach], "RIGHT", -3, 0)
					elseif point == "top" then
						frame[id]:SetPoint("TOPLEFT", frame[attach], "TOP", getHalfWidth(group, position, frame), -3)
					elseif point == "bottom" then
						frame[id]:SetPoint("BOTTOMLEFT", frame[attach], "BOTTOM", getHalfWidth(group, position, frame), 3)
					elseif point == "outright" then
						frame[id]:SetPoint("LEFT", frame[attach], "RIGHT", 3, 0)
					elseif point == "outleft" then
						frame[id]:SetPoint("RIGHT", frame[attach], "LEFT", -3, 0)
					end
				end
			end
		end
	end
end

local getIconSize = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		return ABF_db.profile[group].iconSizes[name]
	end
	return self[name]
end})
local setIconSize = setmetatable({}, {__index=function(self, name)
	self[name] = function(group, value)
		ABF_db.profile[group].iconSizes[name] = value

		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return self[name]
end})

local getPositionArgs
local function iconOptions(name)
	return function(group)
		return {
			type = "group",
			name = L["Position"],
			desc = L["Position of the text."],
			args = getPositionArgs(group, name, true),
		}, {
			type = 'range',
			name = L["Size"],
			desc = L["Size of the icon."],
			get = getIconSize[name],
			set = setIconSize[name],
			passValue = group,
			min = 0.5,
			max = 4,
			step = 0.01,
			bigStep = 0.05,
			isPercent = true,
		}
	end
end

local function createIconHandler(id)
	ABF[id] = iconPosition(id)
	ABF[id .. "_options"] = iconOptions(id)
end

createIconHandler('pvpIcon')
createIconHandler('combatIcon')
createIconHandler('restIcon')
createIconHandler('leaderIcon')
createIconHandler('masterIcon')
createIconHandler('raidTargetIcon')

local barNameToLocal = {
	portrait = L["Portrait"],
	healthBar = L["Health bar"],
	powerBar = L["Power bar"],
	repBar = L["Reputation bar"],
	expBar = L["Experience bar"],
	threatBar = L["Threat bar"],
	castBar = L["Cast bar"],
	blankSpace = L["Blank space"],
}

local barPositionValidate = {}
local function getBarPositionValidate(group, forceUpdate)
	if barPositionValidate[group] then
		if not forceUpdate then
			return barPositionValidate[group]
		end
		for k in pairs(barPositionValidate[group]) do
			barPositionValidate[group][k] = nil
		end
	else
		barPositionValidate[group] = {}
	end
	
	for k,v in pairs(ABF_db.profile[group].barPositions) do
		barPositionValidate[group][v] = k
	end	
	
	local i = 1
	while i <= #barPositionValidate[group] do
		local v = barPositionValidate[group][i]
		local good = false
		if v ~= "portrait" or (ABF_db.profile[group].portraitSide ~= "left" and ABF_db.profile[group].portraitSide ~= "right") then
			for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
				if frame[v] then
					good = true
					break
				end
			end
		end
		if not good then
			table.remove(barPositionValidate[group], i)
			i = i - 1
		else
			barPositionValidate[group][i] = barNameToLocal[v]
		end
		i = i + 1
	end
	
	return barPositionValidate[group]
end

local function createBarOptions_position(id, group, disableFunc)
	local function position_get(group)
		getBarPositionValidate(group, true)
		return barNameToLocal[id]
	end
	local function position_set(group, value)
		local barName
		for k,v in pairs(barNameToLocal) do
			if value == v then
				barName = k
				break
			end
		end
		assert(barName)
		local barPositions = ABF_db.profile[group].barPositions
		local new_position = barPositions[barName]
		local old_position = barPositions[id]
		local step = 1
		if new_position < old_position then
			step = -1
		end
		for k, v in pairs(barPositions) do
			if (new_position <= v and v < old_position) or (old_position < v and v <= new_position) then
				barPositions[k] = v - step
			end
		end
		barPositions[id] = new_position
		getBarPositionValidate(group, true)
		
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return {
		name = L["Position"],
		desc = L["Position of the bar relative to other bars."],
		type = 'text',
		validate = getBarPositionValidate(group),
		passValue = group,
		get = position_get,
		set = position_set,
		disabled = disableFunc,
	}
end

local function barOptions(id)
	local function height_get(group)
		return ABF_db.profile[group].heights[id]
	end
	local function height_set(group, value)
		ABF_db.profile[group].heights[id] = value
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return function(group, disableFunc)
		return {
			name = L["Height"],
			desc = L["Relative height of the bar."],
			type = 'range',
			passValue = group,
			get = height_get,
			set = height_set,
			min = 1,
			max = 12,
			step = 1,
			disabled = disableFunc,
		}, createBarOptions_position(id, group, disableFunc)
	end
end

ABF.healthBar_options = barOptions('healthBar')
ABF.powerBar_options = barOptions('powerBar')
ABF.threatBar_options = barOptions('threatBar')
ABF.castBar_options = barOptions('castBar')
ABF.expBar_options = barOptions('expBar')
ABF.repBar_options = barOptions('repBar')

if playerClass == "DRUID" then
	local function druidManaBarHeight_get(group)
		return ABF_db.profile[group].druidManaBarHeight
	end
	local function druidManaBarHeight_set(group, value)
		ABF_db.profile[group].druidManaBarHeight = value
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	function ABF.druidManaBar_options(group)
		if group == "player" then
			return {
				name = L["Height Percentage"],
				desc = L["Percentage of height that the druid mana bar will take up of the normal power bar's height"],
				type = 'range',
				passValue = group,
				get = druidManaBarHeight_get,
				set = druidManaBarHeight_set,
				min = 0.05,
				max = 0.95,
				step = 0.01,
				bigStep = 0.05,
				isPercent = true,
			}
		end
	end
end

local function portraitSide_get(group)
	return ABF_db.profile[group].portraitSide
end
local function portraitSide_set(group, value)
	ABF_db.profile[group].portraitSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
local function portraitSide_isLeftOrRight(group)
	local portraitSide = ABF_db.profile[group].portraitSide
	return portraitSide == "left" or portraitSide == "right"
end
local function portraitHeight_get(group)
	return ABF_db.profile[group].heights.portrait
end
local function portraitHeight_set(group, value)
	ABF_db.profile[group].heights.portrait = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
function ABF.portrait_options(group)
	local t = createBarOptions_position('portrait', group)
	t.disabled = portraitSide_isLeftOrRight
	return {
		name = L["Side"],
		desc = L["What side of the unit frame to position the portrait at."],
		type = "text",
		validate = {
			left = L["Left"],
			right = L["Right"],
			top = L["Center"],
		},
		get = portraitSide_get,
		set = portraitSide_set,
		passValue = group
	}, 	{
		name = L["Height"],
		desc = L["Relative height of the portrait"],
		type = 'range',
		passValue = group,
		get = portraitHeight_get,
		set = portraitHeight_set,
		min = 1,
		max = 12,
		step = 1,
		disabled = portraitSide_isLeftOrRight,
	}, t
end

local function auraSide_get(group)
	return ABF_db.profile[group].auraSide
end
local function auraSide_set(group, value)
	ABF_db.profile[group].auraSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function buffSide_get(group)
	return ABF_db.profile[group].buffSide
end
local function buffSide_set(group, value)
	ABF_db.profile[group].buffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function debuffSide_get(group)
	return ABF_db.profile[group].debuffSide
end
local function debuffSide_set(group, value)
	ABF_db.profile[group].debuffSide = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end

local function splitOptions_isHidden(group)
	local aura = PitBull:HasModule("Aura") and PitBull:GetModule("Aura")
	if not aura or not aura.db or not aura.db.profile or not aura.db.profile[group] then
		return true
	end
	return not aura.db.profile[group].split
end

local function nonsplitOptions_isHidden(group)
	local aura = PitBull:HasModule("Aura") and PitBull:GetModule("Aura")
	if not aura or not aura.db or not aura.db.profile or not aura.db.profile[group] then
		return true
	end
	return aura.db.profile[group].split
end

function ABF.auraFrame_options(group)
	return {
		name = L["Side"],
		desc = L["Which side to position all aura icons at."],
		type = "text",
		validate = {
			left = "Left",
			topleft = "Top-left",
			bottomleft = "Bottom-left",
			right = "Right",
			topright = "Top-right",
			bottomright = "Bottom-right",
			top = "Top",
			bottom = "Bottom",
		},
		get = auraSide_get,
		set = auraSide_set,
		passValue = group,
		hidden = function(group)
			for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
				if frame.auraFrame2 then
					return true
				end
			end	
			return false
		end,
		disabled = function(group)
			for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
				if frame.auraFrame then
					return false
				end
			end	
			return true
		end,
	}, {
		name = L["Side"],
		desc = L["Options to change the side which the auras are on."],
		type = "group",
		hidden = function()
			for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
				if frame.auraFrame2 then
					return false
				end
			end
			return true
		end,
		passValue = group,
		args = {
			buffside = {
				name = L["Buff side"],
				desc = L["Which side to position buff icons at."],
				type = "text",
				validate = {
					left = "Left",
					topleft = "Top-left",
					bottomleft = "Bottom-left",
					right = "Right",
					topright = "Top-right",
					bottomright = "Bottom-right",
					top = "Top",
					bottom = "Bottom",
				},
				get = buffSide_get,
				set = buffSide_set,
				passValue = group,
				disabled = splitOptions_isHidden,
			},
			debuffside = {
				name = L["Debuff side"],
				desc = L["Which side to position debuff icons at."],
				type = "text",
				validate = {
					left = "Left",
					topleft = "Top-left",
					bottomleft = "Bottom-left",
					right = "Right",
					topright = "Top-right",
					bottomright = "Bottom-right",
					top = "Top",
					bottom = "Bottom",
				},
				get = debuffSide_get,
				set = debuffSide_set,
				passValue = group,
				disabled = splitOptions_isHidden,
			},
		}
	}
end

if playerClass == "ROGUE" or playerClass == "DRUID" then
	local function comboSide_get(group)
		return ABF_db.profile[group].comboSide
	end
	local function comboSide_set(group, value)
		ABF_db.profile[group].comboSide = value
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end

	ABF.comboPoints_options_name = "Combo points"
	ABF.comboPoints_options_desc = "Options for the combo point display."
	function ABF.comboPoints_options(group)
		if group == "target" then
			return {
				name = L["Side"],
				desc = L["What side of the unit frame to position the combo point display at."],
				type = "text",
				validate = {
					left = "Left",
					right = "Right",
					top = "Top",
					bottom = "Bottom",
				},
				get = comboSide_get,
				set = comboSide_set,
				passValue = group,
			}
		end
	end
end

local textStyles = {
	class = {
		["Standard"]            = "[Classification] [Level:DifficultyColor] [SmartClass:ClassColor] [DruidForm:Paren] [SmartRace]",
		["Player Classes Only"] = "[Classification] [Level:DifficultyColor] [PlayerClass:ClassColor] [DruidForm:Paren] [SmartRace]",
		["Short"]               = "[DifficultyColor][Level][Plus][White] [SmartRace]",
	},
	health = {
		["Absolute"]       = "[Status:SureHP:PercentHP:Percent]",
		["Absolute Short"] = "[Status:SureHP:Short:PercentHP:Percent]",
		["Difference"]     = "[Status:SmartMissingHP]",
		["Percent"]        = "[Status:PercentHP:Percent]",
		["Mini"]           = "[CurHP:VeryShort]",
		["Smart"]          = "[IsFriend?Status:MissingHP:HideZero:Short:Color(ff7f7f)!Status:SmartHP:Short]",
		["Absolute and Percent"]  = "[~Status?FractionalHP:Short] || [Status:PercentHP:Percent]",
		["Informational"]  = "[IsFriend?~Status?MissingHP:HideZero:Short:Color(ff7f7f)] || [~Status?FractionalHP:Short] || [Status:PercentHP:Percent]"
	},
	name = {
		["Standard"]             = "[Name] [AFKDND:Angle]",
		["Hostility Colored"]    = "[Name:HostileColor] [AFKDND:Angle]",
		["Class Colored"]        = "[Name:ClassColor] [AFKDND:Angle]",
		["Long"]                 = "[Level] [Name:ClassColor] [AFKDND:Angle]",
		["Long w/ Druid form"]   = "[Level] [Name:ClassColor] [DruidForm:Paren] [AFKDND:Angle]",
	},
	power = {
		["Absolute"]       = "[~HasNoMP?FractionalMP]",
		["Absolute Short"] = "[~HasNoMP?FractionalMP:Short]",
		["Difference"]     = "[MissingMP:Negate]",
		["Percent"]        = "[PercentMP:Percent]",
		["Mini"]           = "[~HasNoMP?CurMP:VeryShort]",
		["Smart"]          = "[MissingMP:HideZero:Short:Color(7f7fff)]",
	},
	threat = {
		["Absolute"]       = "[~HasNoThreat?Threat]",
		["Absolute Short"] = "[~HasNoThreat?Threat:Short]",
		["Fractional"]     = "[~HasNoThreat?FractionalThreat]",
		["Fractional Short"] = "[~HasNoThreat?FractionalThreat:Short]",
		["Difference"]     = "[~HasNoThreat?MissingThreat:Negate]",
		["Percent"]        = "[~HasNoThreat?PercentThreat:Percent]",
		["Mini"]           = "[~HasNoThreat?Threat:VeryShort]",
	},
	combo = {
		["Standard"]       = "[IsEnergy#player?Combos:HideZero]",
	},
	experience = {
		["Standard"]       = "[FractionalXP] [PercentXP:Percent:Paren] [PercentRestXP:HideZero:Percent:Prepend(R: )]",
		["On Mouse-over"]       = "[IsMouseover?FractionalXP] [IsMouseover?PercentXP:Percent:Paren] [IsMouseover?PercentRestXP:HideZero:Percent:Prepend(R: )]",
	},
	reputation = {
		["Standard"]       = "[~IsMouseOver?FractionalRep] [~IsMouseOver?PercentRep:Percent:Paren!RepName]",
	},
}

local styleTypeValidate = {
	class = L["Class"],
	health = L["Health"],
	name = L["Name"],
	power = L["Power"],
	threat = L["Threat"],
	combo = L["Combo points"],
	experience = L["Experience"],
	reputation = L["Reputation"],
	none = L["None"],
}

local styleTypeValidateDesc = {}

local defaultShownTexts = {
	Health = true,
	Class = true,
	Name = true,
	Power = true,
	Threat = function(group) return not group:find("target$") end,
	Combo = function(group) return group == "target" end,
	Experience = function(group) return group == "player" or group == "pet" end,
	Reputation = function(group) return group == "player" end,
}

local function textExtraFrameCreation(id, name, func)
	assert(type(name) == "string")
	return function(unit, frame)
		local shownTexts = ABF_db.profile[frame.group].shownTexts
		if not shownTexts then
			shownTexts = {}
			ABF_db.profile[frame.group].shownTexts = shownTexts
			for k, v in pairs(defaultShownTexts) do
				if type(v) == "function" then
					shownTexts[k] = v(frame.group) or nil
				else
					shownTexts[k] = v or nil
				end
			end
		end
		if func and not func(frame.group) then
			return
		end
		if not shownTexts[name] then
			return
		end
		
		local db = ABF_db.profile[frame.group].texts[name]
		local text = newFrame("FontString", frame.overlay, "OVERLAY")
		frame[id] = text
		local font, fontsize = PitBull:GetFont()
		text:SetFont(font, fontsize * db.size)
		text:SetShadowColor(0, 0, 0, 1)
		text:SetShadowOffset(0.8, -0.8)
		text:SetNonSpaceWrap(false)
		text:SetJustifyH("LEFT")
		local code
		if db.style == "Custom" then
			code = db.custom
		else
			local styles = textStyles[db.styleType]
			code = styles and styles[db.style] or ''
		end	
		if not PitBull.configMode or name == "Name" or code:find("[%[:%?!][Nn][Aa][Mm][Ee][%]:%?!]") then
			DogTag:RegisterFontString(text, frame, unit, code or '')
		else
			text:SetText("[" .. name .. "]")
		end
	end
end

local getTextStyle = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		return ABF_db.profile[group].texts[name].style
	end
	return self[name]
end})
local setTextStyle = setmetatable({}, {__index=function(self, name)
	local id = "layoutText" .. name
	self[name] = function(group, value)
		local db = ABF_db.profile[group].texts[name]
		db.style = value

		local code
		if value == "Custom" then
			code = db.custom
		else
			local styles = textStyles[db.styleType]
			code = styles and styles[value] or ''
			db.custom = code
		end
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			if not PitBull.configMode or name == "Name" or code:find("[%[:%?!][Nn][Aa][Mm][Ee][%]:%?!]") then
				DogTag:RegisterFontString(frame[id], frame, unit, code or '')
			else
				frame[id]:SetText("[" .. name .. "]")
			end
		end
	end
	return self[name]
end})

local getTextCustom = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		return ABF_db.profile[group].texts[name].custom
	end
	return self[name]
end})
local setTextCustom = setmetatable({}, {__index=function(self, name)
	local id = "layoutText" .. name
	self[name] = function(group, value)
		value = DogTag:FixCasing(value)
		local db = ABF_db.profile[group].texts[name]
		db.custom = value

		if db.style ~= "Custom" then
			return
		end
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			if not PitBull.configMode or name == "Name" or value:find("[%[:%?!][Nn][Aa][Mm][Ee][%]:%?!]") then
				DogTag:RegisterFontString(frame[id], frame, unit, value)
			else
				frame[id]:SetText("[" .. name .. "]")
			end
			
		end
	end
	return self[name]
end})

local getTextSize = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		return ABF_db.profile[group].texts[name].size
	end
	return self[name]
end})
local setTextSize = setmetatable({}, {__index=function(self, name)
	self[name] = function(group, value)
		ABF_db.profile[group].texts[name].size = value

		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return self[name]
end})

local getHiddenTextsValidate
local setTextEnabled = setmetatable({}, {__index=function(self, name)
	self[name] = function(group, value)
		ABF_db.profile[group].shownTexts[name] = value or nil
		
		getHiddenTextsValidate(group, true)
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
		
		Dewdrop:Refresh()
	end
	return self[name]
end})

local basePositionValidate = {
	["healthBar-left"] = ("%s/%s"):format(L["Health bar"], L["Left"]),
	["healthBar-center"] = ("%s/%s"):format(L["Health bar"], L["Middle"]),
	["healthBar-right"] = ("%s/%s"):format(L["Health bar"], L["Right"]),
	["healthBar-outright"] = ("%s/%s"):format(L["Health bar"], L["Outside, Right"]),
	["healthBar-outleft"] = ("%s/%s"):format(L["Health bar"], L["Outside, Left"]),
	
	["powerBar-left"] = ("%s/%s"):format(L["Power bar"], L["Left"]),
	["powerBar-center"] = ("%s/%s"):format(L["Power bar"], L["Middle"]),
	["powerBar-right"] = ("%s/%s"):format(L["Power bar"], L["Right"]),
	["powerBar-outright"] = ("%s/%s"):format(L["Power bar"], L["Outside, Right"]),
	["powerBar-outleft"] = ("%s/%s"):format(L["Power bar"], L["Outside, Left"]),
	
	["threatBar-left"] = ("%s/%s"):format(L["Threat bar"], L["Left"]),
	["threatBar-center"] = ("%s/%s"):format(L["Threat bar"], L["Middle"]),
	["threatBar-right"] = ("%s/%s"):format(L["Threat bar"], L["Right"]),
	["threatBar-outright"] = ("%s/%s"):format(L["Threat bar"], L["Outside, Right"]),
	["threatBar-outleft"] = ("%s/%s"):format(L["Threat bar"], L["Outside, Left"]),
	
	["expBar-left"] = ("%s/%s"):format(L["Experience bar"], L["Left"]),
	["expBar-center"] = ("%s/%s"):format(L["Experience bar"], L["Middle"]),
	["expBar-right"] = ("%s/%s"):format(L["Experience bar"], L["Right"]),
	["expBar-outright"] = ("%s/%s"):format(L["Experience bar"], L["Outside, Right"]),
	["expBar-outleft"] = ("%s/%s"):format(L["Experience bar"], L["Outside, Left"]),
	
	["repBar-left"] = ("%s/%s"):format(L["Reputation bar"], L["Left"]),
	["repBar-center"] = ("%s/%s"):format(L["Reputation bar"], L["Middle"]),
	["repBar-right"] = ("%s/%s"):format(L["Reputation bar"], L["Right"]),
	["repBar-outright"] = ("%s/%s"):format(L["Reputation bar"], L["Outside, Right"]),
	["repBar-outleft"] = ("%s/%s"):format(L["Reputation bar"], L["Outside, Left"]),
	
	["blankSpace-left"] = ("%s/%s"):format(L["Blank space"], L["Left"]),
	["blankSpace-center"] = ("%s/%s"):format(L["Blank space"], L["Middle"]),
	["blankSpace-right"] = ("%s/%s"):format(L["Blank space"], L["Right"]),
	["blankSpace-outright"] = ("%s/%s"):format(L["Blank space"], L["Outside, Right"]),
	["blankSpace-outleft"] = ("%s/%s"):format(L["Blank space"], L["Outside, Left"]),
	
	["portrait-top"] = ("%s/%s"):format(L["Portrait"], L["Top"]),
	["portrait-center"] = ("%s/%s"):format(L["Portrait"], L["Middle"]),
	["portrait-bottom"] = ("%s/%s"):format(L["Portrait"], L["Bottom"]),
	
	["frame-outtopleft"] = ("%s/%s"):format(L["Frame"], L["Outside, Above-left"]),
	["frame-outtop"] = ("%s/%s"):format(L["Frame"], L["Outside, Above-middle"]),
	["frame-outtopright"] = ("%s/%s"):format(L["Frame"], L["Outside, Above-right"]),
	["frame-outbottomleft"] = ("%s/%s"):format(L["Frame"], L["Outside, Below-left"]),
	["frame-outbottom"] = ("%s/%s"):format(L["Frame"], L["Outside, Below-middle"]),
	["frame-outbottomright"] = ("%s/%s"):format(L["Frame"], L["Outside, Below-right"]),
	["frame-outlefttop"] = ("%s/%s"):format(L["Frame"], L["Outside, Left-top"]),
	["frame-outleft"] = ("%s/%s"):format(L["Frame"], L["Outside, Left-middle"]),
	["frame-outleftbottom"] = ("%s/%s"):format(L["Frame"], L["Outside, Left-bottom"]),
	["frame-outrighttop"] = ("%s/%s"):format(L["Frame"], L["Outside, Right-top"]),
	["frame-outright"] = ("%s/%s"):format(L["Frame"], L["Outside, Right-middle"]),
	["frame-outrightbottom"] = ("%s/%s"):format(L["Frame"], L["Outside, Right-bottom"]),
	
	["frame-center"] = ("%s/%s"):format(L["Frame"], L["Inside, Center"]),
	["frame-intopleft"] = ("%s/%s"):format(L["Frame"], L["Inside, Top-left"]),
	["frame-intop"] = ("%s/%s"):format(L["Frame"], L["Inside, Top"]),
	["frame-intopright"] = ("%s/%s"):format(L["Frame"], L["Inside, Top-right"]),
	["frame-inbottomleft"] = ("%s/%s"):format(L["Frame"], L["Inside, Bottom-left"]),
	["frame-inbottom"] = ("%s/%s"):format(L["Frame"], L["Inside, Bottom"]),
	["frame-inbottomright"] = ("%s/%s"):format(L["Frame"], L["Inside, Bottom-right"]),
	["frame-inleft"] = ("%s/%s"):format(L["Frame"], L["Inside, Left"]),
	["frame-inright"] = ("%s/%s"):format(L["Frame"], L["Inside, Right"]),
	
	["frame-edgetopleft"] = ("%s/%s"):format(L["Frame"], L["Edge, Top-left"]),
	["frame-edgetop"] = ("%s/%s"):format(L["Frame"], L["Edge, Top"]),
	["frame-edgetopright"] = ("%s/%s"):format(L["Frame"], L["Edge, Top-right"]),
	["frame-edgeleft"] = ("%s/%s"):format(L["Frame"], L["Edge, Left"]),
	["frame-edgeright"] = ("%s/%s"):format(L["Frame"], L["Edge, Right"]),
	["frame-edgebottomleft"] = ("%s/%s"):format(L["Frame"], L["Edge, Bottom-left"]),
	["frame-edgebottom"] = ("%s/%s"):format(L["Frame"], L["Edge, Bottom"]),
	["frame-edgebottomright"] = ("%s/%s"):format(L["Frame"], L["Edge, Bottom-right"]),
	
	["hide"] = L["Hidden"],
}

local setTextPosition = setmetatable({}, {__index=function(self, name)
	self[name] = function(group, value)
		if type(group) == "table" then
			group, value = unpack(group)
		end
		if value ~= "hide" then
			local frame = findTextAtPosition(group, value)
			if frame then
				frame = frame:gsub("^layoutText", "")
				ABF_db.profile[group].texts[frame].position = ABF_db.profile[group].texts[name].position
			end
		end
		ABF_db.profile[group].texts[name].position = value
	
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return self[name]
end})
local getTextPosition = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		if type(group) == "table" then
			local value
			group, value = unpack(group)
			return ABF_db.profile[group].texts[name].position == value
		end
		return ABF_db.profile[group].texts[name].position
	end
	return self[name]
end})

local setIconPosition = setmetatable({}, {__index=function(self, name)
	self[name] = function(group, value)
		if type(group) == "table" then
			group, value = unpack(group)
		end
		ABF_db.profile[group].iconPositions[name] = value
	
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			PitBull:UpdateLayout(frame)
		end
	end
	return self[name]
end})
local getIconPosition = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		if type(group) == "table" then
			local value
			group, value = unpack(group)
			return ABF_db.profile[group].iconPositions[name] == value
		end
		return ABF_db.profile[group].iconPositions[name]
	end
	return self[name]
end})

local getPositionArgs_hidden = setmetatable({}, {__index = function(self, attach)
	self[attach] = function(group)
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			if frame[attach] then
				return false
			end
		end
		return true
	end
	return self[attach]
end})
local positionArgs = {}
function getPositionArgs(group, name, icon)
	if not positionArgs[group] then
		positionArgs[group] = {}
	end
	if positionArgs[group][name] then
		return positionArgs[group][name]
	end
	local t = {}
	positionArgs[group][name] = t
	
	for k, v in pairs(basePositionValidate) do
		local attach, point = ('-'):split(k)
		local attachName, pointName = ('/'):split(v)
		if point then
			if not t[attach] then
				t[attach] = {
					type = 'text',
					name = attachName,
					desc = attachName,
					validate = {},
					get = not icon and getTextPosition[name] or getIconPosition[name],
					set = not icon and setTextPosition[name] or setIconPosition[name],
					passValue = group,
					hidden = attach ~= "frame" and getPositionArgs_hidden[attach]
				}
			end
			t[attach].validate[k] = pointName
		else
			t[attach] = {
				type = 'toggle',
				name = v,
				desc = v,
				get = not icon and getTextPosition[name] or getIconPosition[name],
				set = not icon and setTextPosition[name] or setIconPosition[name],
				passValue = { group, k },
			}
		end
	end
	return t
end

local styleValidates = {}
local styleValidateDescs = {}
local function getStyleValidate(group, name, forceUpdate)
	if not styleValidates[group] then
		styleValidates[group] = {}
		styleValidateDescs[group] = {}
	end
	if styleValidates[group][name] then
		if not forceUpdate then
			return styleValidates[group][name]
		end
		for k in pairs(styleValidates[group][name]) do
			styleValidates[group][name][k] = nil
			styleValidateDescs[group][name][k] = nil
		end
	end
 	local t = styleValidates[group][name] or {}
	local u = styleValidateDescs[group][name] or {}
	styleValidates[group][name] = t
	styleValidateDescs[group][name] = u
	local db = ABF_db.profile[group].texts[name]
	local styleType = db.styleType
	local styles = textStyles[styleType]
	t["Hide"] = L["Hide"]
	u["Hide"] = ""
	t["Custom"] = L["Custom"]
	u["Custom"] = L["Choose your own custom text."]
	if styles then
		for k, v in pairs(styles) do
			t[k] = L[k]
			u[k] = v
		end
	end
	return t
end
local function getStyleValidateDescs(group, name)
	getStyleValidate(group, name) -- to update
	return styleValidateDescs[group][name]
end

local getTextStyleType = setmetatable({}, {__index=function(self, name)
	self[name] = function(group)
		return ABF_db.profile[group].texts[name].styleType
	end
	return self[name]
end})
local setTextStyleType = setmetatable({}, {__index=function(self, name)
	local id = "layoutText" .. name
	self[name] = function(group, value)
		local db = ABF_db.profile[group].texts[name]
		db.styleType = value
		
		getStyleValidate(group, name, true)
		
		if db.style == "Custom" then
			return
		end
		
		local styles = textStyles[db.styleType]
		if not styles then
			db.style = "Custom"
		elseif not styles[db.style] then
			if styles["Standard"] then
				db.style = "Standard"
			elseif styles["Absolute"] then
				db.style = "Absolute"
			else
				db.style = "Custom"
			end
		end
		local code
		if db.style == "Custom" then
			code = db.custom
		else
			code = styles[db.style] or ''
			db.custom = code
		end
		
		for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
			if not PitBull.configMode or name == "Name" or code:find("[%[:%?!][Nn][Aa][Mm][Ee][%]:%?!]") then
				DogTag:RegisterFontString(frame[id], frame, unit, code or '')
			else
				frame[id]:SetText("[" .. name .. "]")
			end
		end
	end
	return self[name]
end})

local function retTrue() return true end
local function retFalse() return false end

local function textOptions(id, name, noStyle)
	assert(type(name) == "string")
	if not noStyle then
		return function(group)
			return {
				type = "text",
				name = L["Style"],
				desc = L["Use a preconfigured style."],
				validate = getStyleValidate(group, name),
				validateDesc = getStyleValidateDescs(group, name),
				get = getTextStyle[name],
				set = setTextStyle[name],
				passValue = group,
			}, {
				type = "text",
				name = L["Style - Custom"],
				desc = L["Create a custom style for this unit type."],
				get = getTextCustom[name],
				set = setTextCustom[name],
				usage = L["DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0"],
				passValue = group,
			}, {
				type = "text",
				name = L["Style - Type"],
				desc = L["Choose the type of pre-made styles to work with."],
				get = getTextStyleType[name],
				set = setTextStyleType[name],
				passValue = group,
				validate = styleTypeValidate,
				validateDesc = styleTypeValidateDesc,
			}, {
				type = "group",
				name = L["Position"],
				desc = L["Position of the text."],
				args = getPositionArgs(group, name),
			}, {
				type = 'range',
				name = L["Size"],
				desc = L["Font size of the text."],
				get = getTextSize[name],
				set = setTextSize[name],
				passValue = group,
				min = 0.25,
				max = 3,
				step = 0.01,
				bigStep = 0.05,
				isPercent = true,
			}, {
				type = 'toggle',
				name = L["Enable"],
				desc = L["Click to disable this text from being shown."],
				get = retTrue,
				set = setTextEnabled[name],
				passValue = group,
			}
		end
	else
		return function(group)
			return {
				type = "group",
				name = L["Position"],
				desc = L["Position of the text."],
				args = getPositionArgs(group, name),
			}
		end
	end
end

local function createArbitraryText(name, extraFrameFunc)
	local textOptionsArg
	if extraFrameFunc == true then
		extraFrameFunc = nil
		textOptionsArg = true
	end
	local id = textOptionsArg and name or "layoutText" .. name
	if ABF[id] then
		return
	end
	ABF[id] = textPosition(id, name)
	if not textOptionsArg then
		ABF.extraFrames[id] = textExtraFrameCreation(id, name, extraFrameFunc)
	end
	-- L[name] should be changed to L:HasTranslation(id) and L[id]
	ABF[id .. "_options_name"] = L["%s text"]:format(L[name])
	ABF[id .. "_options_desc"] = L["Customize the %s text."]:format(L[name])
	ABF[id .. "_options"] = textOptions(id, name, textOptionsArg)
end

createArbitraryText('combatText', true)

local hiddenTextsValidate = {}
function getHiddenTextsValidate(group, forceUpdate)
	if hiddenTextsValidate[group] then
		if not forceUpdate then
			return hiddenTextsValidate[group]
		end
		for k in pairs(hiddenTextsValidate[group]) do
			hiddenTextsValidate[group][k] = nil
		end
	end
	local t = hiddenTextsValidate[group] or {}
	hiddenTextsValidate[group] = t
	local shownTexts = ABF_db.profile[group].shownTexts
	if not shownTexts then
		shownTexts = {}
		ABF_db.profile[group].shownTexts = shownTexts
		for k, v in pairs(defaultShownTexts) do
			if type(v) == "function" then
				shownTexts[k] = v(group) or nil
			else
				shownTexts[k] = v or nil
			end
		end
	end
	for k,v in pairs(defaultShownTexts) do
		if not ABF_db.profile[group].shownTexts[k] then
			t[k] = L[k]
			if type(v) == "function" then
				if not v(group) then
					t[k] = nil
				end
			end
		end
	end
	for k,v in pairs(ABF_db.profile[group].shownTexts) do
		if not v then
			t[k] = L[k]
			if type(defaultShownTexts[k]) == "function" then
				if not defaultShownTexts[k](group) then
					t[k] = nil
				end
			end
		end
	end
	return t
end

local function otherTexts_hidden_set(group, name)
	ABF_db.profile[group].shownTexts[name] = true
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
	getHiddenTextsValidate(group, true)
	Dewdrop:Refresh()
end

local function otherTexts_hidden_disabled(group)
	return not next(hiddenTextsValidate[group])
end

local function otherTexts_new_set(group, name)
	if ABF_db.profile[group].shownTexts[name] then
		return
	end
	ABF_db.profile[group].shownTexts[name] = true
	if ABF_db.profile[group].texts[name].custom == "" then
		ABF_db.profile[group].texts[name].custom = name
	end
	createArbitraryText(name)
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
	Dewdrop:Refresh()
end

function ABF.alwaysText_options(group)
	return {
		type = 'group',
		name = L["Other texts"],
		desc = L["Create other texts."],
		order = 101,
		args = {
			hidden = {
				type = 'text',
				name = L["Currently disabled texts"],
				desc = L["Show some of the currently disabled texts"],
				get = retFalse,
				set = otherTexts_hidden_set,
				validate = getHiddenTextsValidate(group),
				passValue = group,
				disabled = otherTexts_hidden_disabled
			},
			new = {
				type = 'text',
				name = L["New text"],
				desc = L["Create a new text"],
				get = false,
				set = otherTexts_new_set,
				passValue = group,
				usage = L["<Name>"]
			}
		},
	}
end

local function blankSpace_enable_get(group)
	return ABF_db.profile[group].showBlankSpace
end
local function blankSpace_disable_get(group)
	return not ABF_db.profile[group].showBlankSpace
end
local function blankSpace_enable_set(group, value)
	ABF_db.profile[group].showBlankSpace = value
	for unit, frame in PitBull:IterateUnitFramesByGroup(group) do
		PitBull:UpdateLayout(frame)
	end
end
local blackSpace_optionsFunc = barOptions('blankSpace')
function ABF.alwaysBar_options(group)
	return {
		type = 'group',
		name = L["Blank space"],
		desc = L["Change settings for the blank space."],
		args = {
			{
				name = L["Enable"],
				desc = L["Enable the blank space."],
				type = 'toggle',
				passValue = group,
				get = blankSpace_enable_get,
				set = blankSpace_enable_set,
			}, blackSpace_optionsFunc(group, blankSpace_disable_get)
		},
		order = 50,
	}
end

PitBull:RegisterLayout("ABF", ABF)

hooksecurefunc(PitBull, 'OnInitialize', function()
	f = nil
	for k in pairs(defaultShownTexts) do
		createArbitraryText(k)
	end
	for group, data in pairs(ABF_db.profile) do
		if data.shownTexts then
			data.shownTexts.init = nil -- REMOVE LATER
			for k in pairs(data.shownTexts) do
				createArbitraryText(k)
			end
		end
	end
end)